CCCL (CUDA C++ Core Libraries) provides: - CUB: device/block/warp-level GPU primitives (reduce, scan, sort, topk) - Thrust: high-level parallel algorithms (transform_reduce, sort, scan) - libcudacxx: CUDA C++ standard library (atomics, barriers, memory) - cudax: experimental features (memory resources, allocators) - Tuning policies: per-SM hardware-specific algorithm parameters Competition optimization vectors mapped to CCCL: - Output TPS (83% weight): warp_reduce, block_reduce, device_topk - Input TPS (14% weight): device_scan, block_load, prefetch - Cache TPS (3% weight): prefix caching strategy patterns - Memory (0.9 util): pooled/cached/buddy allocators Source: https://github.com/NVIDIA/cccl (shallow clone, HEAD only) License: Apache-2.0
2564 lines
103 KiB
Plaintext
2564 lines
103 KiB
Plaintext
// SPDX-FileCopyrightText: Copyright (c) 2011, Duane Merrill. All rights reserved.
|
|
// SPDX-FileCopyrightText: Copyright (c) 2011-2026, NVIDIA CORPORATION. All rights reserved.
|
|
// SPDX-License-Identifier: BSD-3
|
|
|
|
//! @file
|
|
//! cub::DeviceReduce provides device-wide, parallel operations for computing a reduction across a sequence of data
|
|
//! items residing within device-accessible memory.
|
|
|
|
#pragma once
|
|
|
|
#include <cub/config.cuh>
|
|
|
|
#ifndef CCCL_DISABLE_NVRTC_COMPATIBILITY_CHECK
|
|
# if _CCCL_COMPILER(NVRTC)
|
|
# error \
|
|
"Including <cub/device/device_reduce.cuh> is not supported when compiling with NVRTC. Include block-, warp-, or thread-level primitives instead (e.g. <cub/block/block_reduce.cuh>). You can define CCCL_DISABLE_NVRTC_COMPATIBILITY_CHECK to disable this warning."
|
|
# endif // _CCCL_COMPILER(NVRTC)
|
|
#endif // CCCL_DISABLE_NVRTC_COMPATIBILITY_CHECK
|
|
|
|
#if defined(_CCCL_IMPLICIT_SYSTEM_HEADER_GCC)
|
|
# pragma GCC system_header
|
|
#elif defined(_CCCL_IMPLICIT_SYSTEM_HEADER_CLANG)
|
|
# pragma clang system_header
|
|
#elif defined(_CCCL_IMPLICIT_SYSTEM_HEADER_MSVC)
|
|
# pragma system_header
|
|
#endif // no system header
|
|
|
|
#include <cub/detail/choose_offset.cuh>
|
|
#include <cub/detail/deferred_parameter.cuh>
|
|
#include <cub/detail/device_memory_resource.cuh>
|
|
#include <cub/detail/env_dispatch.cuh>
|
|
#include <cub/detail/temporary_storage.cuh>
|
|
#include <cub/device/dispatch/dispatch_reduce.cuh>
|
|
#include <cub/device/dispatch/dispatch_reduce_by_key.cuh>
|
|
#include <cub/device/dispatch/dispatch_reduce_deterministic.cuh>
|
|
#include <cub/device/dispatch/dispatch_streaming_reduce.cuh>
|
|
#include <cub/thread/thread_operators.cuh>
|
|
#include <cub/util_type.cuh>
|
|
|
|
#include <cuda/__execution/determinism.h>
|
|
#include <cuda/__execution/require.h>
|
|
#include <cuda/__execution/tune.h>
|
|
#include <cuda/__functional/call_or.h>
|
|
#include <cuda/__functional/maximum.h>
|
|
#include <cuda/__functional/minimum.h>
|
|
#include <cuda/__iterator/tabulate_output_iterator.h>
|
|
#include <cuda/__memory_resource/get_memory_resource.h>
|
|
#include <cuda/__stream/get_stream.h>
|
|
#include <cuda/__stream/stream_ref.h>
|
|
#include <cuda/argument>
|
|
#include <cuda/std/__execution/env.h>
|
|
#include <cuda/std/__functional/identity.h>
|
|
#include <cuda/std/__functional/invoke.h>
|
|
#include <cuda/std/__functional/operations.h>
|
|
#include <cuda/std/__iterator/indirectly_comparable.h>
|
|
#include <cuda/std/__type_traits/conditional.h>
|
|
#include <cuda/std/__type_traits/is_integral.h>
|
|
#include <cuda/std/__type_traits/is_same.h>
|
|
#include <cuda/std/__utility/forward.h>
|
|
#include <cuda/std/cstdint>
|
|
#include <cuda/std/limits>
|
|
|
|
CUB_NAMESPACE_BEGIN
|
|
|
|
namespace detail
|
|
{
|
|
template <typename DeterminismT>
|
|
inline constexpr bool is_non_deterministic_v =
|
|
::cuda::std::is_same_v<DeterminismT, ::cuda::execution::determinism::not_guaranteed_t>;
|
|
} // namespace detail
|
|
|
|
//! @rst
|
|
//! DeviceReduce provides device-wide, parallel operations for computing
|
|
//! a reduction across a sequence of data items residing within
|
|
//! device-accessible memory.
|
|
//!
|
|
//! .. image:: ../../img/reduce_logo.png
|
|
//! :align: center
|
|
//!
|
|
//! Overview
|
|
//! ====================================
|
|
//!
|
|
//! A `reduction <http://en.wikipedia.org/wiki/Reduce_(higher-order_function)>`_
|
|
//! (or *fold*) uses a binary combining operator to compute a single aggregate
|
|
//! from a sequence of input elements.
|
|
//!
|
|
//! Usage Considerations
|
|
//! ====================================
|
|
//!
|
|
//! @cdp_class{DeviceReduce}
|
|
//! @determinism{run_to_run}
|
|
//!
|
|
//! Performance
|
|
//! ====================================
|
|
//!
|
|
//! @linear_performance{reduction, reduce-by-key, and run-length encode}
|
|
//!
|
|
//! Tuning
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! All algorithms in DeviceReduce that accept an environment (except ``ReduceByKey``) can be tuned by passing a custom
|
|
//! :ref:`policy selector <cub-policy-selectors>` that returns a :cpp:struct:`cub::ReducePolicy`, as shown in the
|
|
//! example below:
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin reduce-policy-selector
|
|
//! :end-before: example-end reduce-policy-selector
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin reduce-tuning
|
|
//! :end-before: example-end reduce-tuning
|
|
//!
|
|
//! ``DeviceReduce::ReduceByKey`` can be tuned by passing a custom
|
|
//! :ref:`policy selector <cub-policy-selectors>` that returns a :cpp:struct:`cub::ReduceByKeyPolicy`, as shown in the
|
|
//! example below:
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_by_key_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin reduce-by-key-policy-selector
|
|
//! :end-before: example-end reduce-by-key-policy-selector
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_by_key_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin reduce-by-key-tuning
|
|
//! :end-before: example-end reduce-by-key-tuning
|
|
//!
|
|
//! Deferred problem sizes
|
|
//! ====================================
|
|
//!
|
|
//! ``Reduce``, ``Sum``, ``Min``, ``Max``, and ``TransformReduce`` allow specifying the problem size from a value that
|
|
//! resides in device memory through a single-value ``cuda::args::deferred`` argument. The deferred source may be a
|
|
//! device pointer, a one-element span, or a random-access fancy iterator whose element is a non-``bool`` 32- or 64-bit
|
|
//! integer.
|
|
//!
|
|
//! The problem size is read in stream order by the reduction kernels. Work that produces the count in the same stream
|
|
//! is ordered automatically; a producer in another stream requires an event or an equivalent dependency. The source
|
|
//! and its value must remain accessible and unchanged until all reduction kernels complete. The value must be
|
|
//! nonnegative and ``[d_in, d_in + num_items)`` must be accessible. A temporary-storage query does not dereference the
|
|
//! deferred source.
|
|
//!
|
|
//! Deferred reductions are CUDA Graph capturable. The pointed-to count may change between graph replays without
|
|
//! updating or recapturing the graph. Compile-time and runtime bounds are accepted as caller preconditions, but do not
|
|
//! currently change temporary storage, grid dimensions, or pass selection. Since the problem size is not available to
|
|
//! the host, the first reduction pass launches CUB's maximum grid and may launch more blocks than the actual problem
|
|
//! size needs. Extra blocks return without reducing input.
|
|
//!
|
|
//! Determinism
|
|
//! ====================================
|
|
//!
|
|
//! ``cub::DeviceReduce`` supports all three :ref:`determinism guarantees <cccl-determinism>`; the
|
|
//! default is ``run_to_run``.
|
|
//!
|
|
//! - ``run_to_run`` (the default) is reproducible because, for a given GPU, every launch with the same input,
|
|
//! build, and launch configuration selects the *same* tuning policy and therefore performs the *same* fixed
|
|
//! reduction tree: the input is partitioned into the same tiles, mapped onto the same thread blocks, and the
|
|
//! partial results are combined in the same fixed order on every run — no atomics or other run-dependent
|
|
//! ordering are involved. Because floating-point addition is only pseudo-associative, that fixed combining order
|
|
//! is what makes the result bitwise-identical from one run to the next. The combining order is tied to the
|
|
//! tuning and the partition, so it can change on a *different* GPU architecture (or under a different
|
|
//! user-provided tuning); that is exactly the cross-architecture reproducibility that ``gpu_to_gpu`` adds on top.
|
|
//! - ``gpu_to_gpu`` reproducibility depends on the type and operator:
|
|
//!
|
|
//! - ``float`` and ``double`` with ``cuda::std::plus`` use a dedicated, hardware-independent implementation
|
|
//! based on a `Reproducible Floating-point Accumulator (RFA)
|
|
//! <https://people.eecs.berkeley.edu/~demmel/ma221_Fall23/J115_Efficient_Reproducible_Summation_TOMS_2020.pdf>`__:
|
|
//! input values are grouped into a fixed number of exponent-range bins and accumulated in that fixed,
|
|
//! hardware-independent order, so the same inputs yield the same bits on any GPU architecture.
|
|
//! - Exactly-associative cases — integral types with a known CUDA binary operator, and ``float``/``double``
|
|
//! with ``min``/``max`` — are already identical across GPUs, so the request is satisfied by the (faster)
|
|
//! ``run_to_run`` path.
|
|
//! - All other type/operator combinations are rejected at compile time.
|
|
//!
|
|
//! - ``not_guaranteed`` uses an atomic accumulation kernel when the conditions for it are met: a contiguous output
|
|
//! iterator, ``cuda::std::plus``, an accumulator of at least 4 bytes, and an output type equal to the
|
|
//! accumulator type. Atomics combine partial results in whatever order the hardware schedules them, which can
|
|
//! differ between runs — hence no run-to-run guarantee, but typically the fastest option. When those conditions
|
|
//! are not met, the call falls back to ``run_to_run`` rather than failing.
|
|
//!
|
|
//! @endrst
|
|
struct DeviceReduce
|
|
{
|
|
private:
|
|
template <typename EnvT,
|
|
typename InputIteratorT,
|
|
typename OutputIteratorT,
|
|
typename ReductionOpT,
|
|
typename TransformOpT,
|
|
typename T,
|
|
typename NumItemsT,
|
|
::cuda::execution::determinism::__determinism_t Determinism>
|
|
CUB_RUNTIME_FUNCTION static cudaError_t reduce_impl(
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
ReductionOpT reduction_op,
|
|
TransformOpT transform_op,
|
|
T init,
|
|
::cuda::execution::determinism::__determinism_holder_t<Determinism>,
|
|
const EnvT& env)
|
|
{
|
|
using args_traits_t = ::cuda::args::__traits<NumItemsT>;
|
|
using offset_t = detail::choose_offset_t<typename args_traits_t::element_type>;
|
|
using accum_t = decltype(detail::reduce::select_accum_t<InputIteratorT, T, ReductionOpT, TransformOpT>(
|
|
static_cast<detail::use_default*>(nullptr)));
|
|
|
|
if constexpr (Determinism == ::cuda::execution::determinism::__determinism_t::__gpu_to_gpu)
|
|
{
|
|
// Only instantiated with `plus<float|double>`; RFA hardcodes `deterministic_sum_t<accum_t>`.
|
|
(void) reduction_op;
|
|
using default_policy_selector = detail::reduce::
|
|
policy_selector_from_types<accum_t, offset_t, detail::rfa::deterministic_sum_t<accum_t>, Determinism>;
|
|
return detail::dispatch_with_env_and_tuning<default_policy_selector>(
|
|
env, [&](auto policy_selector, void* storage, size_t& bytes, cudaStream_t stream) {
|
|
return detail::rfa::dispatch<InputIteratorT,
|
|
OutputIteratorT,
|
|
decltype(detail::make_num_items_dispatch_arg(num_items)),
|
|
T,
|
|
TransformOpT,
|
|
accum_t>(
|
|
storage,
|
|
bytes,
|
|
d_in,
|
|
d_out,
|
|
detail::make_num_items_dispatch_arg(num_items),
|
|
init,
|
|
stream,
|
|
transform_op,
|
|
policy_selector);
|
|
});
|
|
}
|
|
else if constexpr (Determinism == ::cuda::execution::determinism::__determinism_t::__not_guaranteed)
|
|
{
|
|
using default_policy_selector =
|
|
detail::reduce::policy_selector_from_types<accum_t, offset_t, ReductionOpT, Determinism>;
|
|
return detail::dispatch_with_env_and_tuning<default_policy_selector>(
|
|
env, [&](auto policy_selector, void* storage, size_t& bytes, cudaStream_t stream) {
|
|
return detail::reduce::dispatch<accum_t, /* StableReductionOrder */ false>(
|
|
storage,
|
|
bytes,
|
|
d_in,
|
|
THRUST_NS_QUALIFIER::unwrap_contiguous_iterator(d_out),
|
|
detail::make_num_items_dispatch_arg(num_items),
|
|
reduction_op,
|
|
init,
|
|
stream,
|
|
transform_op,
|
|
policy_selector);
|
|
});
|
|
}
|
|
else
|
|
{
|
|
using default_policy_selector =
|
|
detail::reduce::policy_selector_from_types<accum_t, offset_t, ReductionOpT, Determinism>;
|
|
return detail::dispatch_with_env_and_tuning<default_policy_selector>(
|
|
env, [&](auto policy_selector, void* storage, size_t& bytes, cudaStream_t stream) {
|
|
return detail::reduce::dispatch<accum_t>(
|
|
storage,
|
|
bytes,
|
|
d_in,
|
|
d_out,
|
|
detail::make_num_items_dispatch_arg(num_items),
|
|
reduction_op,
|
|
init,
|
|
stream,
|
|
transform_op,
|
|
policy_selector);
|
|
});
|
|
}
|
|
}
|
|
|
|
//! @brief Internal implementation shared by Reduce and TransformReduce env overloads
|
|
template <typename InputIteratorT,
|
|
typename OutputIteratorT,
|
|
typename ReductionOpT,
|
|
typename TransformOpT,
|
|
typename T,
|
|
typename NumItemsT,
|
|
typename EnvT,
|
|
typename AccumT = decltype(detail::reduce::select_accum_t<InputIteratorT, T, ReductionOpT, TransformOpT>(
|
|
static_cast<detail::use_default*>(nullptr)))>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t __transform_reduce(
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
ReductionOpT reduction_op,
|
|
TransformOpT transform_op,
|
|
T init,
|
|
const EnvT& env)
|
|
{
|
|
static_assert(!::cuda::std::execution::__queryable_with<EnvT, ::cuda::execution::determinism::__get_determinism_t>,
|
|
"Determinism should be used inside requires to have an effect.");
|
|
using requirements_t = ::cuda::std::execution::
|
|
__query_result_or_t<EnvT, ::cuda::execution::__get_requirements_t, ::cuda::std::execution::env<>>;
|
|
using default_determinism_t =
|
|
::cuda::std::execution::__query_result_or_t<requirements_t,
|
|
::cuda::execution::determinism::__get_determinism_t,
|
|
::cuda::execution::determinism::run_to_run_t>;
|
|
|
|
constexpr auto gpu_gpu_determinism =
|
|
::cuda::std::is_same_v<default_determinism_t, ::cuda::execution::determinism::gpu_to_gpu_t>;
|
|
|
|
// integral types are always gpu-to-gpu deterministic if reduction operator is a simple cuda binary
|
|
// operator, so fallback to run-to-run determinism
|
|
constexpr auto integral_fallback =
|
|
gpu_gpu_determinism && ::cuda::std::is_integral_v<AccumT> && (detail::is_cuda_binary_operator<ReductionOpT>);
|
|
|
|
// use gpu-to-gpu determinism only for float and double types with ::cuda::std::plus operator
|
|
constexpr auto float_double_plus =
|
|
gpu_gpu_determinism && detail::is_one_of_v<AccumT, float, double> && detail::is_cuda_std_plus_v<ReductionOpT>;
|
|
|
|
constexpr auto float_double_min_max_fallback =
|
|
gpu_gpu_determinism
|
|
&& detail::is_one_of_v<AccumT, float, double> && detail::is_cuda_minimum_maximum_v<ReductionOpT>;
|
|
|
|
constexpr auto supported =
|
|
integral_fallback || float_double_plus || float_double_min_max_fallback || !gpu_gpu_determinism;
|
|
|
|
// gpu_to_gpu determinism is only supported for integral types with cuda operators, or
|
|
// float and double types with ::cuda::std::plus operator
|
|
static_assert(supported, "gpu_to_gpu determinism is unsupported");
|
|
|
|
if constexpr (!supported)
|
|
{
|
|
return cudaErrorNotSupported;
|
|
}
|
|
else
|
|
{
|
|
constexpr auto no_determinism = detail::is_non_deterministic_v<default_determinism_t>;
|
|
|
|
// Certain conditions must be met to be able to use the non-deterministic
|
|
// kernel. The output iterator must be a contiguous iterator, the reduction
|
|
// operator must be plus (for now), and the output type must match the
|
|
// accumulator type. The non-deterministic kernel atomically accumulates
|
|
// directly into the output, so it cannot preserve AccumT accumulation
|
|
// semantics when the output object has a different type. Additionally, since
|
|
// atomics for types of size < 4B are emulated, they perform poorly, so we fall
|
|
// back to the run-to-run determinism.
|
|
using OutputT = cub::detail::non_void_value_t<OutputIteratorT, cub::detail::it_value_t<InputIteratorT>>;
|
|
|
|
constexpr auto is_contiguous_fallback =
|
|
!no_determinism || THRUST_NS_QUALIFIER::is_contiguous_iterator_v<OutputIteratorT>;
|
|
constexpr auto is_plus_fallback = !no_determinism || detail::is_cuda_std_plus_v<ReductionOpT>;
|
|
constexpr auto is_4b_or_greater = !no_determinism || sizeof(AccumT) >= 4;
|
|
constexpr auto is_output_accum = !no_determinism || ::cuda::std::is_same_v<OutputT, AccumT>;
|
|
|
|
// If the conditions for gpu-to-gpu determinism or non-deterministic
|
|
// reduction are not met, we fall back to run-to-run determinism.
|
|
using determinism_t = ::cuda::std::conditional_t<
|
|
(gpu_gpu_determinism && (integral_fallback || float_double_min_max_fallback))
|
|
|| (no_determinism && !(is_contiguous_fallback && is_plus_fallback && is_4b_or_greater && is_output_accum)),
|
|
::cuda::execution::determinism::run_to_run_t,
|
|
default_determinism_t>;
|
|
|
|
return reduce_impl(d_in, d_out, num_items, reduction_op, transform_op, init, determinism_t{}, env);
|
|
}
|
|
}
|
|
|
|
template <typename InputIteratorT,
|
|
typename OutputIteratorT,
|
|
typename ReductionOpT,
|
|
typename InitValueT,
|
|
typename NumItemsT,
|
|
typename EnvT>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t __minmax_reduce(
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
ReductionOpT reduction_op,
|
|
InitValueT init,
|
|
const EnvT& env)
|
|
{
|
|
static_assert(!::cuda::std::execution::__queryable_with<EnvT, ::cuda::execution::determinism::__get_determinism_t>,
|
|
"Determinism should be used inside requires to have an effect.");
|
|
using requirements_t = ::cuda::std::execution::
|
|
__query_result_or_t<EnvT, ::cuda::execution::__get_requirements_t, ::cuda::std::execution::env<>>;
|
|
using requested_determinism_t =
|
|
::cuda::std::execution::__query_result_or_t<requirements_t,
|
|
::cuda::execution::determinism::__get_determinism_t,
|
|
::cuda::execution::determinism::run_to_run_t>;
|
|
|
|
// Static assert to reject gpu_to_gpu determinism since it's not properly implemented
|
|
static_assert(!::cuda::std::is_same_v<requested_determinism_t, ::cuda::execution::determinism::gpu_to_gpu_t>,
|
|
"gpu_to_gpu determinism is not supported");
|
|
|
|
// TODO(NaderAlAwar): Relax this once non-deterministic implementation for min / max is available
|
|
using determinism_t = ::cuda::execution::determinism::run_to_run_t;
|
|
|
|
return reduce_impl(d_in, d_out, num_items, reduction_op, ::cuda::std::identity{}, init, determinism_t{}, env);
|
|
}
|
|
|
|
public:
|
|
//! @rst
|
|
//! Computes a device-wide reduction using the specified binary ``reduction_op`` functor and initial value ``init``.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - Does not support binary reduction operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates a user-defined min-reduction of a
|
|
//! device vector of ``int`` data elements.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh>
|
|
//! // or equivalently <cub/device/device_reduce.cuh>
|
|
//!
|
|
//! // CustomMin functor
|
|
//! struct CustomMin
|
|
//! {
|
|
//! template <typename T>
|
|
//! __device__ __forceinline__
|
|
//! T operator()(const T &a, const T &b) const {
|
|
//! return (b < a) ? b : a;
|
|
//! }
|
|
//! };
|
|
//!
|
|
//! // Declare, allocate, and initialize device-accessible pointers for
|
|
//! // input and output
|
|
//! int num_items; // e.g., 7
|
|
//! int *d_in; // e.g., [8, 6, 7, 5, 3, 0, 9]
|
|
//! int *d_out; // e.g., [-]
|
|
//! CustomMin min_op;
|
|
//! int init; // e.g., INT_MAX
|
|
//! ...
|
|
//!
|
|
//! // Determine temporary device storage requirements
|
|
//! void *d_temp_storage = nullptr;
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! cub::DeviceReduce::Reduce(
|
|
//! d_temp_storage, temp_storage_bytes,
|
|
//! d_in, d_out, num_items, min_op, init);
|
|
//!
|
|
//! // Allocate temporary storage
|
|
//! cudaMalloc(&d_temp_storage, temp_storage_bytes);
|
|
//!
|
|
//! // Run reduction
|
|
//! cub::DeviceReduce::Reduce(
|
|
//! d_temp_storage, temp_storage_bytes,
|
|
//! d_in, d_out, num_items, min_op, init);
|
|
//!
|
|
//! // d_out <-- [0]
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam ReductionOpT
|
|
//! **[inferred]** Binary reduction functor type having member `T operator()(const T &a, const T &b)`
|
|
//!
|
|
//! @tparam T
|
|
//! **[inferred]** Data element type that is convertible to the `value` type of `InputIteratorT`
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] reduction_op
|
|
//! Binary reduction functor
|
|
//!
|
|
//! @param[in] init
|
|
//! Initial value of the reduction
|
|
//!
|
|
//! @param[in] stream
|
|
//! @rst
|
|
//! **[optional]** CUDA stream to launch kernels within. Default is stream\ :sub:`0`.
|
|
//! @endrst
|
|
template <typename InputIteratorT, typename OutputIteratorT, typename ReductionOpT, typename T, typename NumItemsT>
|
|
CUB_RUNTIME_FUNCTION static cudaError_t Reduce(
|
|
void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
ReductionOpT reduction_op,
|
|
T init,
|
|
cudaStream_t stream = nullptr)
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::Reduce");
|
|
|
|
return detail::reduce::dispatch(
|
|
d_temp_storage,
|
|
temp_storage_bytes,
|
|
d_in,
|
|
d_out,
|
|
detail::make_num_items_dispatch_arg(num_items),
|
|
reduction_op,
|
|
init,
|
|
stream);
|
|
}
|
|
|
|
//! @rst
|
|
//! Computes a device-wide reduction using the specified binary ``reduction_op`` functor and initial value ``init``.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - Does not support binary reduction operators that are non-commutative.
|
|
//! - By default, provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! To request "gpu-to-gpu" determinism, pass ``cuda::execution::require(cuda::execution::determinism::gpu_to_gpu)``
|
|
//! as the `env` parameter.
|
|
//! To request "not-guaranteed" determinism, pass
|
|
//! ``cuda::execution::require(cuda::execution::determinism::not_guaranteed)`` as the `env` parameter.
|
|
//! The non-deterministic implementation is only used when the output type matches the accumulator type.
|
|
//! The accumulator type is the decayed result type of invoking ``reduction_op`` with the initial value and an input
|
|
//! value. For example, reducing ``std::uint8_t`` input with ``cuda::std::plus`` and a ``std::uint8_t`` initial
|
|
//! value accumulates in ``int`` due to integer promotion. If the output type does not match the accumulator type,
|
|
//! CUB falls back to run-to-run determinism.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates a user-defined min-reduction of a
|
|
//! device vector of ``int`` data elements.
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin reduce-env-determinism
|
|
//! :end-before: example-end reduce-env-determinism
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam ReductionOpT
|
|
//! **[inferred]** Binary reduction functor type having member `T operator()(const T &a, const T &b)`
|
|
//!
|
|
//! @tparam T
|
|
//! **[inferred]** Data element type that is convertible to the `value` type of `InputIteratorT`
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is ``cuda::std::execution::env<>``.
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] reduction_op
|
|
//! Binary reduction functor
|
|
//!
|
|
//! @param[in] init
|
|
//! Initial value of the reduction
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is ``cuda::std::execution::env{}``.
|
|
//! @endrst
|
|
template <typename InputIteratorT,
|
|
typename OutputIteratorT,
|
|
typename ReductionOpT,
|
|
typename T,
|
|
typename NumItemsT,
|
|
typename EnvT = ::cuda::std::execution::env<>>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t Reduce(
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
ReductionOpT reduction_op,
|
|
T init,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::Reduce");
|
|
return __transform_reduce(d_in, d_out, num_items, reduction_op, ::cuda::std::identity{}, init, env);
|
|
}
|
|
|
|
//! @rst
|
|
//! Computes a device-wide sum using the addition (``+``) operator.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - Uses ``0`` as the initial value of the reduction.
|
|
//! - Does not support ``+`` operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! To request "gpu-to-gpu" determinism, pass ``cuda::execution::require(cuda::execution::determinism::gpu_to_gpu)``
|
|
//! as the `env` parameter.
|
|
//! To request "not-guaranteed" determinism, pass
|
|
//! ``cuda::execution::require(cuda::execution::determinism::not_guaranteed)`` as the `env` parameter.
|
|
//! The non-deterministic implementation is only used when the output type matches the accumulator type.
|
|
//! The accumulator type is the decayed result type of adding the implicit initial value, whose type is the output
|
|
//! type, to an input value. For example, summing ``std::uint8_t`` input into ``std::uint8_t`` output accumulates in
|
|
//! ``int`` due to integer promotion.
|
|
//! If the output type does not match the accumulator type, CUB falls back to run-to-run determinism.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates a user-defined min-reduction of a
|
|
//! device vector of ``int`` data elements.
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin sum-env-determinism
|
|
//! :end-before: example-end sum-env-determinism
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is `cuda::std::execution::env<>`.
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is `cuda::std::execution::env{}`.
|
|
//! @endrst
|
|
template <typename InputIteratorT,
|
|
typename OutputIteratorT,
|
|
typename NumItemsT,
|
|
typename EnvT = ::cuda::std::execution::env<>>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t
|
|
Sum(InputIteratorT d_in, OutputIteratorT d_out, NumItemsT num_items, const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::Sum");
|
|
using OutputT = cub::detail::non_void_value_t<OutputIteratorT, cub::detail::it_value_t<InputIteratorT>>;
|
|
|
|
return __transform_reduce(d_in, d_out, num_items, ::cuda::std::plus<>{}, ::cuda::std::identity{}, OutputT{}, env);
|
|
}
|
|
|
|
//! @rst
|
|
//! Computes a device-wide sum using the addition (``+``) operator.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - Uses ``0`` as the initial value of the reduction.
|
|
//! - Does not support ``+`` operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the sum-reduction of a device vector
|
|
//! of ``int`` data elements.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh> // or equivalently <cub/device/device_reduce.cuh>
|
|
//!
|
|
//! // Declare, allocate, and initialize device-accessible pointers
|
|
//! // for input and output
|
|
//! int num_items; // e.g., 7
|
|
//! int *d_in; // e.g., [8, 6, 7, 5, 3, 0, 9]
|
|
//! int *d_out; // e.g., [-]
|
|
//! ...
|
|
//!
|
|
//! // Determine temporary device storage requirements
|
|
//! void *d_temp_storage = nullptr;
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! cub::DeviceReduce::Sum(
|
|
//! d_temp_storage, temp_storage_bytes, d_in, d_out, num_items);
|
|
//!
|
|
//! // Allocate temporary storage
|
|
//! cudaMalloc(&d_temp_storage, temp_storage_bytes);
|
|
//!
|
|
//! // Run sum-reduction
|
|
//! cub::DeviceReduce::Sum(d_temp_storage, temp_storage_bytes, d_in, d_out, num_items);
|
|
//!
|
|
//! // d_out <-- [38]
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of `d_in`)
|
|
//!
|
|
//! @param[in] stream
|
|
//! @rst
|
|
//! **[optional]** CUDA stream to launch kernels within. Default is stream\ :sub:`0`.
|
|
//! @endrst
|
|
template <typename InputIteratorT, typename OutputIteratorT, typename NumItemsT>
|
|
CUB_RUNTIME_FUNCTION static cudaError_t
|
|
Sum(void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
cudaStream_t stream = nullptr)
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::Sum");
|
|
|
|
// The output value type
|
|
using OutputT = cub::detail::non_void_value_t<OutputIteratorT, cub::detail::it_value_t<InputIteratorT>>;
|
|
|
|
using init_value_t = OutputT;
|
|
|
|
return detail::reduce::dispatch(
|
|
d_temp_storage,
|
|
temp_storage_bytes,
|
|
d_in,
|
|
d_out,
|
|
detail::make_num_items_dispatch_arg(num_items),
|
|
::cuda::std::plus<>{},
|
|
init_value_t{}, // zero-initialize
|
|
stream);
|
|
}
|
|
|
|
//! @rst
|
|
//! Computes a device-wide minimum using the less-than (``<``) operator.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - Uses ``cuda::std::numeric_limits<T>::max()`` as the initial value of the reduction.
|
|
//! - Does not support ``<`` operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the min-reduction of a device vector of ``int`` data elements.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh>
|
|
//! // or equivalently <cub/device/device_reduce.cuh>
|
|
//!
|
|
//! // Declare, allocate, and initialize device-accessible pointers
|
|
//! // for input and output
|
|
//! int num_items; // e.g., 7
|
|
//! int *d_in; // e.g., [8, 6, 7, 5, 3, 0, 9]
|
|
//! int *d_out; // e.g., [-]
|
|
//! ...
|
|
//!
|
|
//! // Determine temporary device storage requirements
|
|
//! void *d_temp_storage = nullptr;
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! cub::DeviceReduce::Min(
|
|
//! d_temp_storage, temp_storage_bytes, d_in, d_out, num_items);
|
|
//!
|
|
//! // Allocate temporary storage
|
|
//! cudaMalloc(&d_temp_storage, temp_storage_bytes);
|
|
//!
|
|
//! // Run min-reduction
|
|
//! cub::DeviceReduce::Min(
|
|
//! d_temp_storage, temp_storage_bytes, d_in, d_out, num_items);
|
|
//!
|
|
//! // d_out <-- [0]
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] stream
|
|
//! @rst
|
|
//! **[optional]** CUDA stream to launch kernels within. Default is stream\ :sub:`0`.
|
|
//! @endrst
|
|
template <typename InputIteratorT, typename OutputIteratorT, typename NumItemsT>
|
|
CUB_RUNTIME_FUNCTION static cudaError_t
|
|
Min(void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
cudaStream_t stream = nullptr)
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::Min");
|
|
|
|
using InputT = detail::it_value_t<InputIteratorT>;
|
|
using init_value_t = InputT;
|
|
using limits_t = ::cuda::std::numeric_limits<init_value_t>;
|
|
#ifndef CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX
|
|
static_assert(limits_t::is_specialized,
|
|
"cub::DeviceReduce::Min uses cuda::std::numeric_limits<InputIteratorT::value_type>::max() as initial "
|
|
"value, but cuda::std::numeric_limits is not specialized for the iterator's value type. This is "
|
|
"probably a bug and you should specialize cuda::std::numeric_limits. Define "
|
|
"CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX to suppress this check.");
|
|
#endif // CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX
|
|
|
|
return detail::reduce::dispatch(
|
|
d_temp_storage,
|
|
temp_storage_bytes,
|
|
d_in,
|
|
d_out,
|
|
detail::make_num_items_dispatch_arg(num_items),
|
|
::cuda::minimum<>{},
|
|
limits_t::max(),
|
|
stream);
|
|
}
|
|
|
|
//! @rst
|
|
//! Computes a device-wide minimum using the less-than (``<``) operator. The result is written to the output
|
|
//! iterator.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - Uses ``cuda::std::numeric_limits<T>::max()`` as the initial value of the reduction.
|
|
//! - Provides determinism based on the environment's determinism requirements.
|
|
//! To request "run-to-run" determinism, pass ``cuda::execution::require(cuda::execution::determinism::run_to_run)``
|
|
//! as the `env` parameter.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the min-reduction of a device vector of ``int`` data elements.
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin min-env-determinism
|
|
//! :end-before: example-end min-env-determinism
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is `cuda::std::execution::env<>`.
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is `cuda::std::execution::env{}`.
|
|
//! @endrst
|
|
template <typename InputIteratorT,
|
|
typename OutputIteratorT,
|
|
typename NumItemsT,
|
|
typename EnvT = ::cuda::std::execution::env<>>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t
|
|
Min(InputIteratorT d_in, OutputIteratorT d_out, NumItemsT num_items, const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::Min");
|
|
using OutputT = cub::detail::non_void_value_t<OutputIteratorT, cub::detail::it_value_t<InputIteratorT>>;
|
|
using limits_t = ::cuda::std::numeric_limits<OutputT>;
|
|
#ifndef CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX
|
|
static_assert(limits_t::is_specialized,
|
|
"cub::DeviceReduce::Min uses cuda::std::numeric_limits<InputIteratorT::value_type>::max() as initial "
|
|
"value, but cuda::std::numeric_limits is not specialized for the iterator's value type. This is "
|
|
"probably a bug and you should specialize cuda::std::numeric_limits. Define "
|
|
"CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX to suppress this check.");
|
|
#endif // CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX
|
|
return __minmax_reduce(d_in, d_out, num_items, ::cuda::minimum<>{}, limits_t::max(), env);
|
|
}
|
|
|
|
private:
|
|
template <class EnvT>
|
|
[[nodiscard]] _CCCL_API static _CCCL_CONSTEVAL bool __validate_determinism_streaming_reduce() noexcept
|
|
{
|
|
static_assert(!::cuda::std::execution::__queryable_with<EnvT, ::cuda::execution::determinism::__get_determinism_t>,
|
|
"Determinism should be used inside requires to have an effect.");
|
|
using requirements_t = ::cuda::std::execution::
|
|
__query_result_or_t<EnvT, ::cuda::execution::__get_requirements_t, ::cuda::std::execution::env<>>;
|
|
using requested_determinism_t =
|
|
::cuda::std::execution::__query_result_or_t<requirements_t, //
|
|
::cuda::execution::determinism::__get_determinism_t,
|
|
::cuda::execution::determinism::run_to_run_t>;
|
|
// Reject gpu_to_gpu determinism since it's not properly implemented
|
|
return !::cuda::std::is_same_v<requested_determinism_t, ::cuda::execution::determinism::gpu_to_gpu_t>;
|
|
}
|
|
|
|
template <typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename CompareOpT,
|
|
typename EnvT>
|
|
CUB_RUNTIME_FUNCTION static cudaError_t __arg_min(
|
|
void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_min_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
CompareOpT compare_op,
|
|
const EnvT& env)
|
|
{
|
|
static_assert(__validate_determinism_streaming_reduce<EnvT>(), "gpu_to_gpu determinism is not supported");
|
|
|
|
using PerPartitionOffsetT = int; // used by the kernel to index within one partition
|
|
using GlobalOffsetT = ::cuda::std::int64_t; // in the range [d_in, d_in + num_items)
|
|
using reduce_op_t = detail::arg_reduce_op<CompareOpT>;
|
|
|
|
return detail::dispatch_with_env(
|
|
d_temp_storage, temp_storage_bytes, env, [&](auto tuning_env, void* storage, size_t& bytes, auto stream) {
|
|
return detail::reduce::dispatch_streaming_arg_reduce<PerPartitionOffsetT>(
|
|
storage,
|
|
bytes,
|
|
d_in,
|
|
d_min_out,
|
|
d_index_out,
|
|
static_cast<GlobalOffsetT>(num_items),
|
|
reduce_op_t{compare_op},
|
|
stream,
|
|
tuning_env);
|
|
});
|
|
}
|
|
|
|
template <typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename CompareOpT,
|
|
typename EnvT>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t __arg_min(
|
|
InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_min_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
CompareOpT compare_op,
|
|
const EnvT& env)
|
|
{
|
|
static_assert(__validate_determinism_streaming_reduce<EnvT>(), "gpu_to_gpu determinism is not supported");
|
|
|
|
using PerPartitionOffsetT = int; // used by the kernel to index within one partition
|
|
using GlobalOffsetT = ::cuda::std::int64_t; // in the range [d_in, d_in + num_items)
|
|
using reduce_op_t = detail::arg_reduce_op<CompareOpT>;
|
|
|
|
return detail::dispatch_with_env(env, [&](auto tuning_env, void* storage, size_t& bytes, auto stream) {
|
|
return detail::reduce::dispatch_streaming_arg_reduce<PerPartitionOffsetT>(
|
|
storage,
|
|
bytes,
|
|
d_in,
|
|
d_min_out,
|
|
d_index_out,
|
|
static_cast<GlobalOffsetT>(num_items),
|
|
reduce_op_t{compare_op},
|
|
stream,
|
|
tuning_env);
|
|
});
|
|
}
|
|
|
|
public:
|
|
//! @rst
|
|
//! Finds the first device-wide minimum based on a given comparison operator and also returns the index of that item.
|
|
//!
|
|
//! .. versionadded:: 3.4.0
|
|
//! First appears in CUDA Toolkit 13.4.
|
|
//!
|
|
//! - The minimum is written to ``d_min_out``
|
|
//! - The offset of the returned item is written to ``d_index_out``, the offset type being written is of type
|
|
//! ``cuda::std::int64_t``.
|
|
//! - For zero-length inputs, the index ``1`` is written to ``d_index_out`` and, if ``compare_op`` is
|
|
//! ``cuda::std::less`` and ``cuda::std::numeric_limits<T>::is_specialized is ``true``,
|
|
//! ``cuda::std::numeric_limits<T>::max()`` is written to ``d_min_out``, otherwise ``T{}``.
|
|
//! - Does not support comparison operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_min_out`` nor ``d_index_out``.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the argmin-reduction of a device vector
|
|
//! of ``int`` data elements.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh> // or equivalently <cub/device/device_reduce.cuh>
|
|
//! #include <cuda/std/cstdint>
|
|
//!
|
|
//! // Declare, allocate, and initialize device-accessible pointers
|
|
//! // for input and output
|
|
//! int num_items; // e.g., 7
|
|
//! int *d_in; // e.g., [8, 6, -7, 5, 3, 1, -9]
|
|
//! int *d_min_out; // memory for the minimum value
|
|
//! cuda::std::int64_t *d_index_out; // memory for the index of the returned value
|
|
//! ...
|
|
//!
|
|
//! // Define the comparison operator
|
|
//! struct abs_less_t {
|
|
//! template <typename T>
|
|
//! __host__ __device__ bool operator()(const T& a, const T& b) const {
|
|
//! return cuda::std::abs(a) < cuda::std::abs(b);
|
|
//! }
|
|
//! };
|
|
//!
|
|
//! // Determine temporary device storage requirements
|
|
//! void *d_temp_storage = nullptr;
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! cub::DeviceReduce::ArgMin(d_temp_storage, temp_storage_bytes, d_in, d_min_out, d_index_out,
|
|
//! num_items, abs_less_t{});
|
|
//!
|
|
//! // Allocate temporary storage
|
|
//! cudaMalloc(&d_temp_storage, temp_storage_bytes);
|
|
//!
|
|
//! // Run argmin-reduction
|
|
//! cub::DeviceReduce::ArgMin(d_temp_storage, temp_storage_bytes, d_in, d_min_out, d_index_out,
|
|
//! num_items, abs_less_t{});
|
|
//!
|
|
//! // d_min_out <-- 1
|
|
//! // d_index_out <-- 5
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items
|
|
//! (of some type `T`) @iterator
|
|
//!
|
|
//! @tparam ExtremumOutIteratorT
|
|
//! **[inferred]** Output iterator type for recording minimum value
|
|
//!
|
|
//! @tparam IndexOutIteratorT
|
|
//! **[inferred]** Output iterator type for recording index of the returned value
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is ``cuda::std::execution::env<>``.
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Iterator to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_min_out
|
|
//! Iterator to which the minimum value is written
|
|
//!
|
|
//! @param[out] d_index_out
|
|
//! Iterator to which the index of the returned value is written
|
|
//!
|
|
//! @param[in] compare_op
|
|
//! Comparison operator returning ``true`` if the first argument is less than the second
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is ``cuda::std::execution::env{}``.
|
|
//! @endrst
|
|
// TODO(bgruber): this constraint is not accurate, since the implementation will compare the value types of
|
|
// ExtremumOutIteratorT, which is wrong IMO
|
|
_CCCL_TEMPLATE(typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename CompareOpT,
|
|
typename EnvT = ::cuda::std::execution::env<>)
|
|
_CCCL_REQUIRES((::cuda::std::indirectly_comparable<InputIteratorT, InputIteratorT, CompareOpT>) )
|
|
CUB_RUNTIME_FUNCTION static cudaError_t ArgMin(
|
|
void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_min_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
CompareOpT compare_op,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::ArgMin");
|
|
return __arg_min(d_temp_storage, temp_storage_bytes, d_in, d_min_out, d_index_out, num_items, compare_op, env);
|
|
}
|
|
|
|
//! @rst
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//! @endrst
|
|
//!
|
|
//! @overload
|
|
//! @note Uses ``cuda::std::less`` as comparison operator
|
|
_CCCL_TEMPLATE(typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename EnvT = ::cuda::std::execution::env<>)
|
|
_CCCL_REQUIRES((!::cuda::std::indirectly_comparable<InputIteratorT, InputIteratorT, EnvT>) )
|
|
CUB_RUNTIME_FUNCTION static cudaError_t ArgMin(
|
|
void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_min_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
const EnvT& env = {})
|
|
{
|
|
return ArgMin(d_temp_storage, temp_storage_bytes, d_in, d_min_out, d_index_out, num_items, ::cuda::std::less{}, env);
|
|
}
|
|
|
|
public:
|
|
//! @rst
|
|
//! Finds the first device-wide minimum using the less-than (``<``) operator and also returns the index of that item.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - The minimum is written to ``d_min_out``
|
|
//! - The offset of the returned item is written to ``d_index_out``, the offset type being written is of type
|
|
//! ``cuda::std::int64_t``.
|
|
//! - For zero-length inputs, the index ``1`` is written to ``d_index_out`` and, if ``compare_op`` is
|
|
//! ``cuda::std::less`` and ``cuda::std::numeric_limits<T>::is_specialized is ``true``,
|
|
//! ``cuda::std::numeric_limits<T>::max()`` is written to ``d_min_out``, otherwise ``T{}``.
|
|
//! - Does not support ``<`` operators that are non-commutative.
|
|
//! - Provides determinism based on the environment's determinism requirements.
|
|
//! To request "run-to-run" determinism, pass ``cuda::execution::require(cuda::execution::determinism::run_to_run)``
|
|
//! as the `env` parameter.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_min_out`` nor ``d_index_out``.
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the argmin-reduction of a device vector of ``int`` data elements.
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin argmin-env-determinism
|
|
//! :end-before: example-end argmin-env-determinism
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items
|
|
//! (of some type `T`) @iterator
|
|
//!
|
|
//! @tparam ExtremumOutIteratorT
|
|
//! **[inferred]** Output iterator type for recording minimum value
|
|
//!
|
|
//! @tparam IndexOutIteratorT
|
|
//! **[inferred]** Output iterator type for recording index of the returned value
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is ``cuda::std::execution::env<>``.
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Iterator to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_min_out
|
|
//! Iterator to which the minimum value is written
|
|
//!
|
|
//! @param[out] d_index_out
|
|
//! Iterator to which the index of the returned value is written
|
|
//!
|
|
//! @param[in] compare_op
|
|
//! Comparison operator returning ``true`` if the first argument is less than the second
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is ``cuda::std::execution::env{}``.
|
|
//! @endrst
|
|
// TODO(bgruber): this constraint is not accurate, since the implementation will compare the value types of
|
|
// ExtremumOutIteratorT, which is wrong IMO
|
|
_CCCL_TEMPLATE(typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename CompareOpT,
|
|
typename EnvT = ::cuda::std::execution::env<>)
|
|
_CCCL_REQUIRES((::cuda::std::indirectly_comparable<InputIteratorT, InputIteratorT, CompareOpT>) )
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t ArgMin(
|
|
InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_min_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
CompareOpT compare_op,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::ArgMin");
|
|
return __arg_min(d_in, d_min_out, d_index_out, num_items, compare_op, env);
|
|
}
|
|
|
|
//! @overload
|
|
//! @note Uses ``cuda::std::less`` as comparison operator
|
|
// TODO(bgruber): this constraint is not accurate, since the implementation will compare the value types of
|
|
// ExtremumOutIteratorT, which is wrong IMO
|
|
_CCCL_TEMPLATE(typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename EnvT = ::cuda::std::execution::env<>)
|
|
_CCCL_REQUIRES((!::cuda::std::indirectly_comparable<InputIteratorT, InputIteratorT, EnvT>) )
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t ArgMin(
|
|
InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_min_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::ArgMin");
|
|
return __arg_min(d_in, d_min_out, d_index_out, num_items, ::cuda::std::less{}, env);
|
|
}
|
|
|
|
//! @rst
|
|
//! Finds the first device-wide minimum using the less-than (``<``) operator, also returning the index of that item.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - The output value type of ``d_out`` is ``cub::KeyValuePair<int, T>``
|
|
//! (assuming the value type of ``d_in`` is ``T``)
|
|
//!
|
|
//! - The minimum is written to ``d_out.value`` and its offset in the input array is written to ``d_out.key``.
|
|
//! - The ``{1, cuda::std::numeric_limits<T>::max()}`` tuple is produced for zero-length inputs
|
|
//!
|
|
//! - Does not support ``<`` operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the argmin-reduction of a device vector
|
|
//! of ``int`` data elements.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh> // or equivalently <cub/device/device_reduce.cuh>
|
|
//!
|
|
//! // Declare, allocate, and initialize device-accessible pointers
|
|
//! // for input and output
|
|
//! int num_items; // e.g., 7
|
|
//! int *d_in; // e.g., [8, 6, 7, 5, 3, 0, 9]
|
|
//! KeyValuePair<int, int> *d_argmin; // e.g., [{-,-}]
|
|
//! ...
|
|
//!
|
|
//! // Determine temporary device storage requirements
|
|
//! void *d_temp_storage = nullptr;
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! cub::DeviceReduce::ArgMin(d_temp_storage, temp_storage_bytes, d_in, d_argmin, num_items);
|
|
//!
|
|
//! // Allocate temporary storage
|
|
//! cudaMalloc(&d_temp_storage, temp_storage_bytes);
|
|
//!
|
|
//! // Run argmin-reduction
|
|
//! cub::DeviceReduce::ArgMin(d_temp_storage, temp_storage_bytes, d_in, d_argmin, num_items);
|
|
//!
|
|
//! // d_argmin <-- [{5, 0}]
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items
|
|
//! (of some type `T`) @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate
|
|
//! (having value type ``cub::KeyValuePair<int, T>``) @iterator
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] stream
|
|
//! @rst
|
|
//! **[optional]** CUDA stream to launch kernels within. Default is stream\ :sub:`0`.
|
|
//! @endrst
|
|
template <typename InputIteratorT, typename OutputIteratorT>
|
|
CCCL_DEPRECATED_BECAUSE("CUB has superseded this interface in favor of the ArgMin interface that takes two separate "
|
|
"iterators: one iterator to which the extremum is written and another iterator to which the "
|
|
"index of the found extremum is written. ") CUB_RUNTIME_FUNCTION static cudaError_t
|
|
ArgMin(void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
int num_items,
|
|
cudaStream_t stream = nullptr)
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::ArgMin");
|
|
|
|
// Signed integer type for global offsets
|
|
using OffsetT = int;
|
|
|
|
// The input type
|
|
using InputValueT = cub::detail::it_value_t<InputIteratorT>;
|
|
|
|
// The output tuple type
|
|
using OutputTupleT = cub::detail::non_void_value_t<OutputIteratorT, KeyValuePair<OffsetT, InputValueT>>;
|
|
|
|
using AccumT = OutputTupleT;
|
|
|
|
using init_value_t = detail::reduce::empty_problem_init_t<AccumT>;
|
|
|
|
// The output value type
|
|
using OutputValueT = typename OutputTupleT::Value;
|
|
|
|
// Wrapped input iterator to produce index-value <OffsetT, InputT> tuples
|
|
using ArgIndexInputIteratorT = ArgIndexInputIterator<InputIteratorT, OffsetT, OutputValueT>;
|
|
|
|
ArgIndexInputIteratorT d_indexed_in(d_in);
|
|
|
|
// Initial value
|
|
init_value_t initial_value{AccumT(1, ::cuda::std::numeric_limits<InputValueT>::max())};
|
|
|
|
return detail::reduce::dispatch<AccumT>(
|
|
d_temp_storage,
|
|
temp_storage_bytes,
|
|
d_indexed_in,
|
|
d_out,
|
|
OffsetT{num_items},
|
|
detail::arg_min{},
|
|
initial_value,
|
|
stream);
|
|
}
|
|
|
|
//! @rst
|
|
//! Computes a device-wide maximum using the greater-than (``>``) operator.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - Uses ``cuda::std::numeric_limits<T>::lowest()`` as the initial value of the reduction.
|
|
//! - Does not support ``>`` operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the max-reduction of a device vector of ``int`` data elements.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh> // or equivalently <cub/device/device_reduce.cuh>
|
|
//!
|
|
//! // Declare, allocate, and initialize device-accessible pointers
|
|
//! // for input and output
|
|
//! int num_items; // e.g., 7
|
|
//! int *d_in; // e.g., [8, 6, 7, 5, 3, 0, 9]
|
|
//! int *d_max; // e.g., [-]
|
|
//! ...
|
|
//!
|
|
//! // Determine temporary device storage requirements
|
|
//! void *d_temp_storage = nullptr;
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! cub::DeviceReduce::Max(d_temp_storage, temp_storage_bytes, d_in, d_max, num_items);
|
|
//!
|
|
//! // Allocate temporary storage
|
|
//! cudaMalloc(&d_temp_storage, temp_storage_bytes);
|
|
//!
|
|
//! // Run max-reduction
|
|
//! cub::DeviceReduce::Max(d_temp_storage, temp_storage_bytes, d_in, d_max, num_items);
|
|
//!
|
|
//! // d_max <-- [9]
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] stream
|
|
//! @rst
|
|
//! **[optional]** CUDA stream to launch kernels within. Default is stream\ :sub:`0`.
|
|
//! @endrst
|
|
template <typename InputIteratorT, typename OutputIteratorT, typename NumItemsT>
|
|
CUB_RUNTIME_FUNCTION static cudaError_t
|
|
Max(void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
cudaStream_t stream = nullptr)
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::Max");
|
|
|
|
using InputT = detail::it_value_t<InputIteratorT>;
|
|
using init_value_t = InputT;
|
|
using limits_t = ::cuda::std::numeric_limits<init_value_t>;
|
|
#ifndef CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX
|
|
static_assert(limits_t::is_specialized,
|
|
"cub::DeviceReduce::Max uses cuda::std::numeric_limits<InputIteratorT::value_type>::lowest() as "
|
|
"initial value, but cuda::std::numeric_limits is not specialized for the iterator's value type. This "
|
|
"is probably a bug and you should specialize cuda::std::numeric_limits. Define "
|
|
"CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX to suppress this check.");
|
|
#endif // CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX
|
|
|
|
return detail::reduce::dispatch(
|
|
d_temp_storage,
|
|
temp_storage_bytes,
|
|
d_in,
|
|
d_out,
|
|
detail::make_num_items_dispatch_arg(num_items),
|
|
::cuda::maximum<>{},
|
|
limits_t::lowest(),
|
|
stream);
|
|
}
|
|
|
|
//! @rst
|
|
//! Computes a device-wide maximum using the greater-than (``>``) operator. The result is written to the output
|
|
//! iterator.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - Uses ``cuda::std::numeric_limits<T>::lowest()`` as the initial value of the reduction.
|
|
//! - Provides determinism based on the environment's determinism requirements.
|
|
//! To request "run-to-run" determinism, pass ``cuda::execution::require(cuda::execution::determinism::run_to_run)``
|
|
//! as the `env` parameter.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the max-reduction of a device vector of ``int`` data elements.
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin max-env-determinism
|
|
//! :end-before: example-end max-env-determinism
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is `cuda::std::execution::env<>`.
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is ``cuda::std::execution::env{}``.
|
|
//! @endrst
|
|
template <typename InputIteratorT,
|
|
typename OutputIteratorT,
|
|
typename NumItemsT,
|
|
typename EnvT = ::cuda::std::execution::env<>>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t
|
|
Max(InputIteratorT d_in, OutputIteratorT d_out, NumItemsT num_items, const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::Max");
|
|
using OutputT = cub::detail::non_void_value_t<OutputIteratorT, cub::detail::it_value_t<InputIteratorT>>;
|
|
using limits_t = ::cuda::std::numeric_limits<OutputT>;
|
|
#ifndef CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX
|
|
static_assert(
|
|
limits_t::is_specialized,
|
|
"cub::DeviceReduce::Max uses cuda::std::numeric_limits<InputIteratorT::value_type>::lowest() as initial value, "
|
|
"but cuda::std::numeric_limits is not specialized for the iterator's value type. This is probably a bug and you "
|
|
"should specialize cuda::std::numeric_limits. Define "
|
|
"CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX to suppress this check.");
|
|
#endif // CCCL_SUPPRESS_NUMERIC_LIMITS_CHECK_IN_CUB_DEVICE_REDUCE_MIN_MAX
|
|
return __minmax_reduce(d_in, d_out, num_items, ::cuda::maximum<>{}, limits_t::lowest(), env);
|
|
}
|
|
|
|
//! @rst
|
|
//! Finds the first device-wide maximum based on a given comparison operator and also returns the index of that item.
|
|
//!
|
|
//! .. versionadded:: 3.4.0
|
|
//! First appears in CUDA Toolkit 13.4.
|
|
//!
|
|
//! - The maximum is written to ``d_max_out``
|
|
//! - The offset of the returned item is written to ``d_index_out``, the offset type being written is of type
|
|
//! ``cuda::std::int64_t``.
|
|
//! - For zero-length inputs, the index ``1`` is written to ``d_index_out`` and, if ``compare_op`` is
|
|
//! ``cuda::std::less`` and ``cuda::std::numeric_limits<T>::is_specialized is ``true``,
|
|
//! ``cuda::std::numeric_limits<T>::lowest()`` is written to ``d_min_out``, otherwise ``T{}``.
|
|
//! - Does not support ``>`` operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the argmax-reduction of a device vector
|
|
//! of `int` data elements.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh> // or equivalently <cub/device/device_reduce.cuh>
|
|
//! #include <cuda/std/cstdint>
|
|
//!
|
|
//! // Declare, allocate, and initialize device-accessible pointers
|
|
//! // for input and output
|
|
//! int num_items; // e.g., 7
|
|
//! int *d_in; // e.g., [8, 6, -7, 5, 3, 1, -9]
|
|
//! int *d_max_out; // memory for the maximum value
|
|
//! cuda::std::int64_t *d_index_out; // memory for the index of the returned value
|
|
//! ...
|
|
//!
|
|
//! // Define the comparison operator
|
|
//! struct abs_less_t {
|
|
//! template <typename T>
|
|
//! __host__ __device__ bool operator()(const T& a, const T& b) const {
|
|
//! return cuda::std::abs(a) < cuda::std::abs(b);
|
|
//! }
|
|
//! };
|
|
//!
|
|
//! // Determine temporary device storage requirements
|
|
//! void *d_temp_storage = nullptr;
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! cub::DeviceReduce::ArgMax(
|
|
//! d_temp_storage, temp_storage_bytes, d_in, d_max_out, d_index_out, num_items, abs_less_t{});
|
|
//!
|
|
//! // Allocate temporary storage
|
|
//! cudaMalloc(&d_temp_storage, temp_storage_bytes);
|
|
//!
|
|
//! // Run argmax-reduction
|
|
//! cub::DeviceReduce::ArgMax(
|
|
//! d_temp_storage, temp_storage_bytes, d_in, d_max_out, d_index_out, num_items, abs_less_t{});
|
|
//!
|
|
//! // d_max_out <-- -9
|
|
//! // d_index_out <-- 6
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items (of some type `T`) @iterator
|
|
//!
|
|
//! @tparam ExtremumOutIteratorT
|
|
//! **[inferred]** Output iterator type for recording maximum value
|
|
//!
|
|
//! @tparam IndexOutIteratorT
|
|
//! **[inferred]** Output iterator type for recording index of the returned value
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is ``cuda::std::execution::env<>``.
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_max_out
|
|
//! Iterator to which the maximum value is written
|
|
//!
|
|
//! @param[out] d_index_out
|
|
//! Iterator to which the index of the returned value is written
|
|
//!
|
|
//! @param[in] compare_op
|
|
//! Comparison operator returning ``true`` if the first argument is less than the second
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is ``cuda::std::execution::env{}``.
|
|
//! @endrst
|
|
// TODO(bgruber): this constraint is not accurate, since the implementation will compare the value types of
|
|
// ExtremumOutIteratorT, which is wrong IMO
|
|
_CCCL_TEMPLATE(typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename CompareOpT,
|
|
typename EnvT = ::cuda::std::execution::env<>)
|
|
_CCCL_REQUIRES((::cuda::std::indirectly_comparable<InputIteratorT, InputIteratorT, CompareOpT>) )
|
|
CUB_RUNTIME_FUNCTION static cudaError_t ArgMax(
|
|
void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_max_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
CompareOpT compare_op,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::ArgMax");
|
|
return __arg_min(
|
|
d_temp_storage, temp_storage_bytes, d_in, d_max_out, d_index_out, num_items, detail::swap_args{compare_op}, env);
|
|
}
|
|
|
|
//! @rst
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//! @overload
|
|
//! @note Uses ``cuda::std::less`` as comparison operator
|
|
//! @endrst
|
|
_CCCL_TEMPLATE(typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename EnvT = ::cuda::std::execution::env<>)
|
|
_CCCL_REQUIRES((!::cuda::std::indirectly_comparable<InputIteratorT, InputIteratorT, EnvT>) )
|
|
CUB_RUNTIME_FUNCTION static cudaError_t ArgMax(
|
|
void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_max_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::ArgMax");
|
|
return __arg_min(
|
|
d_temp_storage, temp_storage_bytes, d_in, d_max_out, d_index_out, num_items, ::cuda::std::greater{}, env);
|
|
}
|
|
|
|
//! @rst
|
|
//! Finds the first device-wide maximum using the greater-than (``>``)
|
|
//! operator, also returning the index of that item
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - The output value type of ``d_out`` is ``cub::KeyValuePair<int, T>``
|
|
//! (assuming the value type of ``d_in`` is ``T``)
|
|
//!
|
|
//! - The maximum is written to ``d_out.value`` and its offset in the input
|
|
//! array is written to ``d_out.key``.
|
|
//! - The ``{1, cuda::std::numeric_limits<T>::lowest()}`` tuple is produced for zero-length inputs
|
|
//!
|
|
//! - Does not support ``>`` operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the argmax-reduction of a device vector
|
|
//! of `int` data elements.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh>
|
|
//! // or equivalently <cub/device/device_reduce.cuh>
|
|
//!
|
|
//! // Declare, allocate, and initialize device-accessible pointers
|
|
//! // for input and output
|
|
//! int num_items; // e.g., 7
|
|
//! int *d_in; // e.g., [8, 6, 7, 5, 3, 0, 9]
|
|
//! KeyValuePair<int, int> *d_argmax; // e.g., [{-,-}]
|
|
//! ...
|
|
//!
|
|
//! // Determine temporary device storage requirements
|
|
//! void *d_temp_storage = nullptr;
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! cub::DeviceReduce::ArgMax(
|
|
//! d_temp_storage, temp_storage_bytes, d_in, d_argmax, num_items);
|
|
//!
|
|
//! // Allocate temporary storage
|
|
//! cudaMalloc(&d_temp_storage, temp_storage_bytes);
|
|
//!
|
|
//! // Run argmax-reduction
|
|
//! cub::DeviceReduce::ArgMax(
|
|
//! d_temp_storage, temp_storage_bytes, d_in, d_argmax, num_items);
|
|
//!
|
|
//! // d_argmax <-- [{6, 9}]
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items (of some type `T`) @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate
|
|
//! (having value type `cub::KeyValuePair<int, T>`) @iterator
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] stream
|
|
//! @rst
|
|
//! **[optional]** CUDA stream to launch kernels within. Default is stream\ :sub:`0`.
|
|
//! @endrst
|
|
template <typename InputIteratorT, typename OutputIteratorT>
|
|
CCCL_DEPRECATED_BECAUSE("CUB has superseded this interface in favor of the ArgMax interface that takes two separate "
|
|
"iterators: one iterator to which the extremum is written and another iterator to which the "
|
|
"index of the found extremum is written. ") CUB_RUNTIME_FUNCTION static cudaError_t
|
|
ArgMax(void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
int num_items,
|
|
cudaStream_t stream = nullptr)
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::ArgMax");
|
|
|
|
// Signed integer type for global offsets
|
|
using OffsetT = int;
|
|
|
|
// The input type
|
|
using InputValueT = cub::detail::it_value_t<InputIteratorT>;
|
|
|
|
// The output tuple type
|
|
using OutputTupleT = cub::detail::non_void_value_t<OutputIteratorT, KeyValuePair<OffsetT, InputValueT>>;
|
|
|
|
using AccumT = OutputTupleT;
|
|
|
|
// The output value type
|
|
using OutputValueT = typename OutputTupleT::Value;
|
|
|
|
using init_value_t = detail::reduce::empty_problem_init_t<AccumT>;
|
|
|
|
// Wrapped input iterator to produce index-value <OffsetT, InputT> tuples
|
|
using ArgIndexInputIteratorT = ArgIndexInputIterator<InputIteratorT, OffsetT, OutputValueT>;
|
|
|
|
ArgIndexInputIteratorT d_indexed_in(d_in);
|
|
|
|
// Initial value
|
|
init_value_t initial_value{AccumT(1, ::cuda::std::numeric_limits<InputValueT>::lowest())};
|
|
|
|
return detail::reduce::dispatch<AccumT>(
|
|
d_temp_storage,
|
|
temp_storage_bytes,
|
|
d_indexed_in,
|
|
d_out,
|
|
OffsetT{num_items},
|
|
detail::arg_max{},
|
|
initial_value,
|
|
stream);
|
|
}
|
|
|
|
//! @rst
|
|
//! Finds the first device-wide maximum using the greater-than (``>``) operator and also returns the index of that
|
|
//! item.
|
|
//!
|
|
//! .. versionadded:: 3.4.0
|
|
//! First appears in CUDA Toolkit 13.4.
|
|
//!
|
|
//! - The maximum is written to ``d_max_out``
|
|
//! - The offset of the returned item is written to ``d_index_out``, the offset type being written is of type
|
|
//! ``cuda::std::int64_t``.
|
|
//! - For zero-length inputs, the index ``1`` is written to ``d_index_out`` and, if ``compare_op`` is
|
|
//! ``cuda::std::less`` and ``cuda::std::numeric_limits<T>::is_specialized is ``true``,
|
|
//! ``cuda::std::numeric_limits<T>::lowest()`` is written to ``d_min_out``, otherwise ``T{}``.
|
|
//! - Does not support ``>`` operators that are non-commutative.
|
|
//! - Provides determinism based on the environment's determinism requirements.
|
|
//! To request "run-to-run" determinism, pass ``cuda::execution::require(cuda::execution::determinism::run_to_run)``
|
|
//! as the `env` parameter.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_max_out`` nor ``d_index_out``.
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the argmax-reduction of a device vector of ``int`` data elements.
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin argmax-env-determinism
|
|
//! :end-before: example-end argmax-env-determinism
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items
|
|
//! (of some type `T`) @iterator
|
|
//!
|
|
//! @tparam ExtremumOutIteratorT
|
|
//! **[inferred]** Output iterator type for recording maximum value
|
|
//!
|
|
//! @tparam IndexOutIteratorT
|
|
//! **[inferred]** Output iterator type for recording index of the returned value
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is ``cuda::std::execution::env<>``.
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Iterator to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_max_out
|
|
//! Iterator to which the maximum value is written
|
|
//!
|
|
//! @param[out] d_index_out
|
|
//! Iterator to which the index of the returned value is written
|
|
//!
|
|
//! @param[in] compare_op
|
|
//! Comparison operator returning ``true`` if the first argument is less than the second
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is ``cuda::std::execution::env{}``.
|
|
//! @endrst
|
|
// TODO(bgruber): this constraint is not accurate, since the implementation will compare the value types of
|
|
// ExtremumOutIteratorT, which is wrong IMO
|
|
_CCCL_TEMPLATE(typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename CompareOpT,
|
|
typename EnvT = ::cuda::std::execution::env<>)
|
|
_CCCL_REQUIRES((::cuda::std::indirectly_comparable<InputIteratorT, InputIteratorT, CompareOpT>) )
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t ArgMax(
|
|
InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_max_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
CompareOpT compare_op,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::ArgMax");
|
|
return __arg_min(d_in, d_max_out, d_index_out, num_items, detail::swap_args{compare_op}, env);
|
|
}
|
|
|
|
//! @overload
|
|
//! @note Uses ``cuda::std::less`` as comparison operator
|
|
template <typename InputIteratorT,
|
|
typename ExtremumOutIteratorT,
|
|
typename IndexOutIteratorT,
|
|
typename EnvT = ::cuda::std::execution::env<>,
|
|
// TODO(bgruber): this constraint is not accurate, since the implementation will compare the value types of
|
|
// ExtremumOutIteratorT, which is wrong IMO
|
|
::cuda::std::enable_if_t<!::cuda::std::indirectly_comparable<InputIteratorT, InputIteratorT, EnvT>, int> = 0>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t
|
|
ArgMax(InputIteratorT d_in,
|
|
ExtremumOutIteratorT d_max_out,
|
|
IndexOutIteratorT d_index_out,
|
|
::cuda::std::int64_t num_items,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::ArgMax");
|
|
return __arg_min(d_in, d_max_out, d_index_out, num_items, ::cuda::std::greater{}, env);
|
|
}
|
|
|
|
//! @rst
|
|
//! Fuses transform and reduce operations
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! - Does not support binary reduction operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates a user-defined min-reduction of a
|
|
//! device vector of `int` data elements.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh>
|
|
//! // or equivalently <cub/device/device_reduce.cuh>
|
|
//!
|
|
//! thrust::device_vector<int> in = { 1, 2, 3, 4 };
|
|
//! thrust::device_vector<int> out(1);
|
|
//!
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! uint8_t *d_temp_storage = nullptr;
|
|
//!
|
|
//! const int init = 42;
|
|
//!
|
|
//! cub::DeviceReduce::TransformReduce(
|
|
//! d_temp_storage,
|
|
//! temp_storage_bytes,
|
|
//! in.begin(),
|
|
//! out.begin(),
|
|
//! in.size(),
|
|
//! cuda::std::plus<>{},
|
|
//! square_t{},
|
|
//! init);
|
|
//!
|
|
//! thrust::device_vector<uint8_t> temp_storage(temp_storage_bytes);
|
|
//! d_temp_storage = temp_storage.data().get();
|
|
//!
|
|
//! cub::DeviceReduce::TransformReduce(
|
|
//! d_temp_storage,
|
|
//! temp_storage_bytes,
|
|
//! in.begin(),
|
|
//! out.begin(),
|
|
//! in.size(),
|
|
//! cuda::std::plus<>{},
|
|
//! square_t{},
|
|
//! init);
|
|
//!
|
|
//! // out[0] <-- 72
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam ReductionOpT
|
|
//! **[inferred]** Binary reduction functor type having member `T operator()(const T &a, const T &b)`
|
|
//!
|
|
//! @tparam TransformOpT
|
|
//! **[inferred]** Unary reduction functor type having member `auto operator()(const T &a)`
|
|
//!
|
|
//! @tparam T
|
|
//! **[inferred]** Data element type that is convertible to the `value` type of `InputIteratorT`
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] reduction_op
|
|
//! Binary reduction functor
|
|
//!
|
|
//! @param[in] transform_op
|
|
//! Unary transform functor
|
|
//!
|
|
//! @param[in] init
|
|
//! Initial value of the reduction
|
|
//!
|
|
//! @param[in] stream
|
|
//! @rst
|
|
//! **[optional]** CUDA stream to launch kernels within. Default is stream\ :sub:`0`.
|
|
//! @endrst
|
|
template <typename InputIteratorT,
|
|
typename OutputIteratorT,
|
|
typename ReductionOpT,
|
|
typename TransformOpT,
|
|
typename T,
|
|
typename NumItemsT>
|
|
CUB_RUNTIME_FUNCTION static cudaError_t TransformReduce(
|
|
void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
ReductionOpT reduction_op,
|
|
TransformOpT transform_op,
|
|
T init,
|
|
cudaStream_t stream = nullptr)
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::TransformReduce");
|
|
|
|
return detail::reduce::dispatch(
|
|
d_temp_storage,
|
|
temp_storage_bytes,
|
|
d_in,
|
|
d_out,
|
|
detail::make_num_items_dispatch_arg(num_items),
|
|
reduction_op,
|
|
init,
|
|
stream,
|
|
transform_op);
|
|
}
|
|
|
|
//! @rst
|
|
//! Computes a device-wide reduction using the specified binary ``reduction_op`` functor,
|
|
//! a unary ``transform_op`` functor, and initial value ``init``.
|
|
//!
|
|
//! .. versionadded:: 3.4.0
|
|
//! First appears in CUDA Toolkit 13.4.
|
|
//!
|
|
//! - Does not support binary reduction operators that are non-commutative.
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! To request "gpu-to-gpu" determinism, pass ``cuda::execution::require(cuda::execution::determinism::gpu_to_gpu)``
|
|
//! as the `env` parameter.
|
|
//! To request "not-guaranteed" determinism, pass
|
|
//! ``cuda::execution::require(cuda::execution::determinism::not_guaranteed)`` as the `env` parameter.
|
|
//! The non-deterministic implementation is only used when the output type matches the accumulator type.
|
|
//! The accumulator type is the decayed result type of invoking ``reduction_op`` with the initial value and the
|
|
//! transformed input value. For example, reducing transformed ``std::uint8_t`` values with ``cuda::std::plus`` and
|
|
//! a
|
|
//! ``std::uint8_t`` initial value accumulates in ``int`` due to integer promotion.
|
|
//! If the output type does not match the accumulator type, CUB falls back to run-to-run determinism.
|
|
//! - The range ``[d_in, d_in + num_items)`` shall not overlap ``d_out``.
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates a transform-reduction with determinism
|
|
//! of a device vector of ``int`` data elements.
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin transform-reduce-env-determinism
|
|
//! :end-before: example-end transform-reduce-env-determinism
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam InputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input items @iterator
|
|
//!
|
|
//! @tparam OutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the reduced aggregate @iterator
|
|
//!
|
|
//! @tparam ReductionOpT
|
|
//! **[inferred]** Binary reduction functor type having member `T operator()(const T &a, const T &b)`
|
|
//!
|
|
//! @tparam TransformOpT
|
|
//! **[inferred]** Unary reduction functor type having member `auto operator()(const T &a)`
|
|
//!
|
|
//! @tparam T
|
|
//! **[inferred]** Data element type that is convertible to the `value` type of `InputIteratorT`
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is ``cuda::std::execution::env<>``.
|
|
//! Supports customization of stream via ``cuda::get_stream``.
|
|
//!
|
|
//! @param[in] d_in
|
|
//! Pointer to the input sequence of data items
|
|
//!
|
|
//! @param[out] d_out
|
|
//! Pointer to the output aggregate
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of input items (i.e., length of ``d_in``)
|
|
//!
|
|
//! @param[in] reduction_op
|
|
//! Binary reduction functor
|
|
//!
|
|
//! @param[in] transform_op
|
|
//! Unary transform functor
|
|
//!
|
|
//! @param[in] init
|
|
//! Initial value of the reduction
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is ``cuda::std::execution::env{}``.
|
|
//! @endrst
|
|
template <typename InputIteratorT,
|
|
typename OutputIteratorT,
|
|
typename ReductionOpT,
|
|
typename TransformOpT,
|
|
typename T,
|
|
typename NumItemsT,
|
|
typename EnvT = ::cuda::std::execution::env<>>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t TransformReduce(
|
|
InputIteratorT d_in,
|
|
OutputIteratorT d_out,
|
|
NumItemsT num_items,
|
|
ReductionOpT reduction_op,
|
|
TransformOpT transform_op,
|
|
T init,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::TransformReduce");
|
|
return __transform_reduce(d_in, d_out, num_items, reduction_op, transform_op, init, env);
|
|
}
|
|
|
|
//! @rst
|
|
//! Reduces segments of values, where segments are demarcated by corresponding runs of identical keys.
|
|
//!
|
|
//! .. versionadded:: 3.4.0
|
|
//! First appears in CUDA Toolkit 13.4.
|
|
//!
|
|
//! This operation computes segmented reductions within ``d_values_in`` using the specified binary ``reduction_op``
|
|
//! functor. The segments are identified by "runs" of corresponding keys in `d_keys_in`, where runs are maximal
|
|
//! ranges of consecutive, identical keys. For the *i*\ :sup:`th` run encountered, the last key of the run and
|
|
//! the corresponding value aggregate of that run are written to ``d_unique_out[i]`` and ``d_aggregates_out[i]``,
|
|
//! respectively. The total number of runs encountered is written to ``d_num_runs_out``.
|
|
//!
|
|
//! - The ``==`` equality operator is used to determine whether keys are equivalent
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! - Let ``out`` be any of
|
|
//! ``[d_unique_out, d_unique_out + *d_num_runs_out)``
|
|
//! ``[d_aggregates_out, d_aggregates_out + *d_num_runs_out)``
|
|
//! ``d_num_runs_out``. The ranges represented by ``out`` shall not overlap
|
|
//! ``[d_keys_in, d_keys_in + num_items)``,
|
|
//! ``[d_values_in, d_values_in + num_items)`` nor ``out`` in any way.
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates a reduce-by-key operation
|
|
//! of a device vector of ``int`` keys and values.
|
|
//!
|
|
//! .. literalinclude:: ../../../cub/test/catch2_test_device_reduce_env_api.cu
|
|
//! :language: c++
|
|
//! :dedent:
|
|
//! :start-after: example-begin reduce-by-key-env
|
|
//! :end-before: example-end reduce-by-key-env
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam KeysInputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input keys @iterator
|
|
//!
|
|
//! @tparam UniqueOutputIteratorT
|
|
//! **[inferred]** Random-access output iterator type for writing unique output keys @iterator
|
|
//!
|
|
//! @tparam ValuesInputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input values @iterator
|
|
//!
|
|
//! @tparam AggregatesOutputIteratorT
|
|
//! **[inferred]** Random-access output iterator type for writing output value aggregates @iterator
|
|
//!
|
|
//! @tparam NumRunsOutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the number of runs encountered @iterator
|
|
//!
|
|
//! @tparam ReductionOpT
|
|
//! **[inferred]** Binary reduction functor type having member `T operator()(const T &a, const T &b)`
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @tparam EnvT
|
|
//! **[inferred]** Execution environment type. Default is ``cuda::std::execution::env<>``.
|
|
//! Supports customization of stream via ``cuda::get_stream``.
|
|
//!
|
|
//! @param[in] d_keys_in
|
|
//! Pointer to the input sequence of keys
|
|
//!
|
|
//! @param[out] d_unique_out
|
|
//! Pointer to the output sequence of unique keys (one key per run)
|
|
//!
|
|
//! @param[in] d_values_in
|
|
//! Pointer to the input sequence of corresponding values
|
|
//!
|
|
//! @param[out] d_aggregates_out
|
|
//! Pointer to the output sequence of value aggregates
|
|
//! (one aggregate per run)
|
|
//!
|
|
//! @param[out] d_num_runs_out
|
|
//! Pointer to total number of runs encountered
|
|
//! (i.e., the length of ``d_unique_out``)
|
|
//!
|
|
//! @param[in] reduction_op
|
|
//! Binary reduction functor
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of associated key+value pairs
|
|
//! (i.e., the length of ``d_in_keys`` and ``d_in_values``)
|
|
//!
|
|
//! @param[in] env
|
|
//! @rst
|
|
//! **[optional]** Execution environment. Default is ``cuda::std::execution::env{}``.
|
|
//! @endrst
|
|
template <typename KeysInputIteratorT,
|
|
typename UniqueOutputIteratorT,
|
|
typename ValuesInputIteratorT,
|
|
typename AggregatesOutputIteratorT,
|
|
typename NumRunsOutputIteratorT,
|
|
typename ReductionOpT,
|
|
typename NumItemsT,
|
|
typename EnvT = ::cuda::std::execution::env<>>
|
|
[[nodiscard]] CUB_RUNTIME_FUNCTION static cudaError_t ReduceByKey(
|
|
KeysInputIteratorT d_keys_in,
|
|
UniqueOutputIteratorT d_unique_out,
|
|
ValuesInputIteratorT d_values_in,
|
|
AggregatesOutputIteratorT d_aggregates_out,
|
|
NumRunsOutputIteratorT d_num_runs_out,
|
|
ReductionOpT reduction_op,
|
|
NumItemsT num_items,
|
|
const EnvT& env = {})
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE("cub::DeviceReduce::ReduceByKey");
|
|
|
|
using OffsetT = detail::choose_offset_t<NumItemsT>;
|
|
using EqualityOp = ::cuda::std::equal_to<>;
|
|
using default_policy_selector = detail::reduce_by_key::policy_selector_from_types<
|
|
ReductionOpT,
|
|
::cuda::std::__accumulator_t<ReductionOpT, detail::it_value_t<ValuesInputIteratorT>>,
|
|
detail::non_void_value_t<UniqueOutputIteratorT, detail::it_value_t<KeysInputIteratorT>>>;
|
|
return detail::dispatch_with_env_and_tuning<default_policy_selector>(
|
|
env, [&](auto policy_selector, void* storage, size_t& bytes, cudaStream_t stream) {
|
|
return detail::reduce_by_key::dispatch(
|
|
storage,
|
|
bytes,
|
|
d_keys_in,
|
|
d_unique_out,
|
|
d_values_in,
|
|
d_aggregates_out,
|
|
d_num_runs_out,
|
|
EqualityOp{},
|
|
reduction_op,
|
|
static_cast<OffsetT>(num_items),
|
|
stream,
|
|
policy_selector);
|
|
});
|
|
}
|
|
|
|
//! @rst
|
|
//! Reduces segments of values, where segments are demarcated by corresponding runs of identical keys.
|
|
//!
|
|
//! .. versionadded:: 2.2.0
|
|
//! First appears in CUDA Toolkit 12.3.
|
|
//!
|
|
//! This operation computes segmented reductions within ``d_values_in`` using the specified binary ``reduction_op``
|
|
//! functor. The segments are identified by "runs" of corresponding keys in `d_keys_in`, where runs are maximal
|
|
//! ranges of consecutive, identical keys. For the *i*\ :sup:`th` run encountered, the last key of the run and
|
|
//! the corresponding value aggregate of that run are written to ``d_unique_out[i]`` and ``d_aggregates_out[i]``,
|
|
//! respectively. The total number of runs encountered is written to ``d_num_runs_out``.
|
|
//!
|
|
//! - The ``==`` equality operator is used to determine whether keys are equivalent
|
|
//! - Provides "run-to-run" determinism for pseudo-associative reduction
|
|
//! (e.g., addition of floating point types) on the same GPU device.
|
|
//! However, results for pseudo-associative reduction may be inconsistent
|
|
//! from one device to a another device of a different compute-capability
|
|
//! because CUB can employ different tile-sizing for different architectures.
|
|
//! - Let ``out`` be any of
|
|
//! ``[d_unique_out, d_unique_out + *d_num_runs_out)``
|
|
//! ``[d_aggregates_out, d_aggregates_out + *d_num_runs_out)``
|
|
//! ``d_num_runs_out``. The ranges represented by ``out`` shall not overlap
|
|
//! ``[d_keys_in, d_keys_in + num_items)``,
|
|
//! ``[d_values_in, d_values_in + num_items)`` nor ``out`` in any way.
|
|
//! - @devicestorage
|
|
//!
|
|
//! Snippet
|
|
//! +++++++++++++++++++++++++++++++++++++++++++++
|
|
//!
|
|
//! The code snippet below illustrates the segmented reduction of ``int`` values grouped by runs of
|
|
//! associated ``int`` keys.
|
|
//!
|
|
//! .. code-block:: c++
|
|
//!
|
|
//! #include <cub/cub.cuh>
|
|
//! // or equivalently <cub/device/device_reduce.cuh>
|
|
//!
|
|
//! // CustomMin functor
|
|
//! struct CustomMin
|
|
//! {
|
|
//! template <typename T>
|
|
//! __device__ __forceinline__
|
|
//! T operator()(const T &a, const T &b) const {
|
|
//! return (b < a) ? b : a;
|
|
//! }
|
|
//! };
|
|
//!
|
|
//! // Declare, allocate, and initialize device-accessible pointers
|
|
//! // for input and output
|
|
//! int num_items; // e.g., 8
|
|
//! int *d_keys_in; // e.g., [0, 2, 2, 9, 5, 5, 5, 8]
|
|
//! int *d_values_in; // e.g., [0, 7, 1, 6, 2, 5, 3, 4]
|
|
//! int *d_unique_out; // e.g., [-, -, -, -, -, -, -, -]
|
|
//! int *d_aggregates_out; // e.g., [-, -, -, -, -, -, -, -]
|
|
//! int *d_num_runs_out; // e.g., [-]
|
|
//! CustomMin reduction_op;
|
|
//! ...
|
|
//!
|
|
//! // Determine temporary device storage requirements
|
|
//! void *d_temp_storage = nullptr;
|
|
//! size_t temp_storage_bytes = 0;
|
|
//! cub::DeviceReduce::ReduceByKey(
|
|
//! d_temp_storage, temp_storage_bytes,
|
|
//! d_keys_in, d_unique_out, d_values_in,
|
|
//! d_aggregates_out, d_num_runs_out, reduction_op, num_items);
|
|
//!
|
|
//! // Allocate temporary storage
|
|
//! cudaMalloc(&d_temp_storage, temp_storage_bytes);
|
|
//!
|
|
//! // Run reduce-by-key
|
|
//! cub::DeviceReduce::ReduceByKey(
|
|
//! d_temp_storage, temp_storage_bytes,
|
|
//! d_keys_in, d_unique_out, d_values_in,
|
|
//! d_aggregates_out, d_num_runs_out, reduction_op, num_items);
|
|
//!
|
|
//! // d_unique_out <-- [0, 2, 9, 5, 8]
|
|
//! // d_aggregates_out <-- [0, 1, 6, 2, 4]
|
|
//! // d_num_runs_out <-- [5]
|
|
//!
|
|
//! @endrst
|
|
//!
|
|
//! @tparam KeysInputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input keys @iterator
|
|
//!
|
|
//! @tparam UniqueOutputIteratorT
|
|
//! **[inferred]** Random-access output iterator type for writing unique output keys @iterator
|
|
//!
|
|
//! @tparam ValuesInputIteratorT
|
|
//! **[inferred]** Random-access input iterator type for reading input values @iterator
|
|
//!
|
|
//! @tparam AggregatesOutputIterator
|
|
//! **[inferred]** Random-access output iterator type for writing output value aggregates @iterator
|
|
//!
|
|
//! @tparam NumRunsOutputIteratorT
|
|
//! **[inferred]** Output iterator type for recording the number of runs encountered @iterator
|
|
//!
|
|
//! @tparam ReductionOpT
|
|
//! **[inferred]** Binary reduction functor type having member `T operator()(const T &a, const T &b)`
|
|
//!
|
|
//! @tparam NumItemsT
|
|
//! **[inferred]** Type of num_items
|
|
//!
|
|
//! @param[in] d_temp_storage
|
|
//! @devicestorage
|
|
//!
|
|
//! @param[in,out] temp_storage_bytes
|
|
//! Reference to size in bytes of ``d_temp_storage`` allocation
|
|
//!
|
|
//! @param[in] d_keys_in
|
|
//! Pointer to the input sequence of keys
|
|
//!
|
|
//! @param[out] d_unique_out
|
|
//! Pointer to the output sequence of unique keys (one key per run)
|
|
//!
|
|
//! @param[in] d_values_in
|
|
//! Pointer to the input sequence of corresponding values
|
|
//!
|
|
//! @param[out] d_aggregates_out
|
|
//! Pointer to the output sequence of value aggregates
|
|
//! (one aggregate per run)
|
|
//!
|
|
//! @param[out] d_num_runs_out
|
|
//! Pointer to total number of runs encountered
|
|
//! (i.e., the length of ``d_unique_out``)
|
|
//!
|
|
//! @param[in] reduction_op
|
|
//! Binary reduction functor
|
|
//!
|
|
//! @param[in] num_items
|
|
//! Total number of associated key+value pairs
|
|
//! (i.e., the length of ``d_in_keys`` and ``d_in_values``)
|
|
//!
|
|
//! @param[in] stream
|
|
//! @rst
|
|
//! **[optional]** CUDA stream to launch kernels within. Default is stream\ :sub:`0`.
|
|
//! @endrst
|
|
template <typename KeysInputIteratorT,
|
|
typename UniqueOutputIteratorT,
|
|
typename ValuesInputIteratorT,
|
|
typename AggregatesOutputIteratorT,
|
|
typename NumRunsOutputIteratorT,
|
|
typename ReductionOpT,
|
|
typename NumItemsT>
|
|
CUB_RUNTIME_FUNCTION _CCCL_FORCEINLINE static cudaError_t ReduceByKey(
|
|
void* d_temp_storage,
|
|
size_t& temp_storage_bytes,
|
|
KeysInputIteratorT d_keys_in,
|
|
UniqueOutputIteratorT d_unique_out,
|
|
ValuesInputIteratorT d_values_in,
|
|
AggregatesOutputIteratorT d_aggregates_out,
|
|
NumRunsOutputIteratorT d_num_runs_out,
|
|
ReductionOpT reduction_op,
|
|
NumItemsT num_items,
|
|
cudaStream_t stream = nullptr)
|
|
{
|
|
_CCCL_NVTX_RANGE_SCOPE_IF(d_temp_storage, "cub::DeviceReduce::ReduceByKey");
|
|
|
|
using OffsetT = detail::choose_offset_t<NumItemsT>;
|
|
using EqualityOp = ::cuda::std::equal_to<>;
|
|
|
|
return detail::reduce_by_key::dispatch(
|
|
d_temp_storage,
|
|
temp_storage_bytes,
|
|
d_keys_in,
|
|
d_unique_out,
|
|
d_values_in,
|
|
d_aggregates_out,
|
|
d_num_runs_out,
|
|
EqualityOp{},
|
|
reduction_op,
|
|
static_cast<OffsetT>(num_items),
|
|
stream);
|
|
}
|
|
};
|
|
CUB_NAMESPACE_END
|